@@ -18,6 +18,7 @@
  */
 
 #include "manager.h"
+#include "downloader.h"
 
 namespace kiwix
 {
@@ -88,7 +89,7 @@ bool Manager::parseXmlDom(const pugi::xml_document& doc,
   return true;
 }
 
-bool Manager::readXml(const string xml,
+bool Manager::readXml(const string& xml,
                       const bool readOnly,
                       const string libraryPath)
 {
@@ -103,6 +104,67 @@ bool Manager::readXml(const string xml,
   return true;
 }
 
+
+
+bool Manager::parseOpdsDom(const pugi::xml_document& doc, const std::string& urlHost)
+{
+  pugi::xml_node libraryNode = doc.child("feed");
+
+  for (pugi::xml_node entryNode = libraryNode.child("entry"); entryNode;
+       entryNode = entryNode.next_sibling("entry")) {
+    kiwix::Book book;
+
+    book.readOnly = false;
+    book.id = entryNode.child("id").child_value();
+    book.title = entryNode.child("title").child_value();
+    book.description = entryNode.child("summary").child_value();
+    book.language = entryNode.child("language").child_value();
+    book.date = entryNode.child("updated").child_value();
+    book.creator = entryNode.child("author").child("name").child_value();
+    for(pugi::xml_node linkNode = entryNode.child("link"); linkNode;
+        linkNode = linkNode.next_sibling("link")) {
+       std::string rel = linkNode.attribute("rel").value();
+
+       if (rel == "http://opds-spec.org/image/thumbnail") {
+         auto faviconUrl = urlHost + linkNode.attribute("href").value();
+         auto downloader = Downloader();
+         auto fileHandle = downloader.download(faviconUrl);
+         if (fileHandle.success) {
+           auto content = getFileContent(fileHandle.path);
+           book.favicon = base64_encode((const unsigned char*)content.data(), content.size());
+           book.faviconMimeType = linkNode.attribute("type").value();
+         } else {
+           std::cerr << "Cannot get favicon content from " << faviconUrl << std::endl;
+         }
+
+       } else if (rel == "http://opds-spec.org/acquisition/open-access") {
+         book.url = linkNode.attribute("href").value();
+       }
+    }
+
+    /* Update the book properties with the new importer */
+    library.addBook(book);
+  }
+
+  return true;
+}
+
+
+
+bool Manager::readOpds(const string& content, const std::string& urlHost)
+{
+  pugi::xml_document doc;
+  pugi::xml_parse_result result
+      = doc.load_buffer_inplace((void*)content.data(), content.size());
+
+  if (result) {
+    this->parseOpdsDom(doc, urlHost);
+    return true;
+  }
+
+  return false;
+}
+
 bool Manager::readFile(const string path, const bool readOnly)
 {
   return this->readFile(path, path, readOnly);
@@ -231,6 +293,7 @@ bool Manager::writeFile(const string path)
   return true;
 }
 
+
 bool Manager::setCurrentBookId(const string id)
 {
   if (library.current.empty() || library.current.top() != id) {
@@ -243,7 +306,7 @@ bool Manager::setCurrentBookId(const string id)
   return true;
 }
 
-string Manager::getCurrentBookId()
+string Manager::getCurrentBookId() const
 {
   return library.current.empty() ? "" : library.current.top();
 }
@@ -495,11 +558,6 @@ bool Manager::setBookIndex(const string id,
   return false;
 }
 
-bool Manager::setBookIndex(const string id, const string path)
-{
-  return this->setBookIndex(id, path, XAPIAN);
-}
-
 bool Manager::setBookPath(const string id, const string path)
 {
   std::vector<kiwix::Book>::iterator itr;
@@ -630,6 +688,24 @@ bool Manager::listBooks(const supportedListMode mode,
   return true;
 }
 
+
+Library Manager::filter(const std::string& search) {
+  Library library;
+
+  if (search.empty()) {
+    return library;
+  }
+
+  for(auto book:this->library.books) {
+     if (matchRegex(book.title, "\\Q" + search + "\\E")
+         || matchRegex(book.description, "\\Q" + search + "\\E")) {
+       library.addBook(book);
+     }
+  }
+
+  return library;
+}
+
 void Manager::checkAndCleanBookPaths(Book& book, const string& libraryPath)
 {
   if (!book.path.empty()) {
